==> autonomize_autorag-0.1.11-py3-none-any.whl: autorag/embedding/modelhub.py <==
"""Modelhub Embedding module implementation"""

# pylint: disable=line-too-long

import base64
import json
import time
from typing import Any, List, Literal, Optional, Union

import httpx

from autorag.embedding.base import Embedding
from autorag.utilities.logger import get_logger

logger = get_logger()


class ModelhubEmbedding(Embedding):
    """
    Modelhub Embedding implementation.

    This class provides an implementation of the Embedding abstract class
    using our internal Modelhub SDK's embedding models to generate embeddings.

    Example:
    .. code-block:: python

        from autorag.embedding import ModelhubEmbedding

        model_name = "bge_base"
        model = ModelhubEmbedding(
            client_id="modelhub-client",
            client_secret="xxx"
        )

        model.create_embedding("Sample text")
    """

    AUTH_URL = "https://auth.sprint.autonomize.dev/realms/autonomize/protocol/openid-connect/token"
    MODEL_URL = (
        "https://clinical-llm.modelhub.sprint.autonomize.dev/v1/model/embedding/predict"
    )

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[str] = None,
    ):
        """
        Initialize the ModelhubEmbedding instance.

        Args:
            client_id (Optional[str]): The client ID for authentication.
            client_secret (Optional[str]): The client secret for authentication.
            token (Optional[str]): An existing JWT token for authentication.

        Raises:
            ValueError: If neither (client_id and client_secret) nor token is provided.
        """

        if client_id is None or client_secret is None:
            if token is None:
                raise ValueError(
                    "Either (`client_id` and `client_secret`) or `token` must be provided."
                )
            logger.warning(
                "It is recommended to provide `client_id` and `client_secret` over `token` "
                "because token gets expired within 24 hours. "
                "You can use `ModelhubEmbedding.is_token_expired(your_token)` to check if it has expired or not."
            )

        self._client_id = client_id
        self._client_secret = client_secret
        self._token = token

    @staticmethod
    def is_token_expired(token: str) -> bool:
        """
        Checks whether the provided JWT token is expired or not.

        Args:
            token (str): The JWT token to check if its expired or not.

        Returns:
            bool: True if token has expired, otherwise False.

        Raises:
            ValueError: When the token provided is not base64
        """

        # Split the token by `.` to get it in list
        b64_components = token.split(".")

        # The token should have three components: header, payload and signature.
        if len(b64_components) != 3:
            raise ValueError("Ill-formatted token provided. Please recheck your token.")

        # Divide it into, header, payload and signature,
        # but we don't need header and signature.
        _, payload_b64, _ = b64_components

        payload_json = base64.urlsafe_b64decode(payload_b64 + "==").decode("utf-8")
        payload = json.loads(payload_json)

        # Get expiry time in Unix-epoch time.
        token_expiry_time = payload["exp"]

        return token_expiry_time < int(time.time())

    def _get_token(self) -> str:
        """
        Obtains a JWT token for authorization if token is not provided or has expired.

        Returns:
            str: JWT token.
        """
        if not self._client_id or not self._client_secret:
            raise ValueError(
                "client_id and client_secret must be provided to fetch JWT token."
            )

        with httpx.Client() as client:
            response = client.post(
                self.AUTH_URL,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data={
                    "grant_type": "client_credentials",
                    "client_id": self._client_id,
                    "client_secret": self._client_secret,
                    "scope": "openid",
                },
            )

            if response.status_code == 401:
                raise ValueError(
                    "Client Error `401 Unauthorized`: client_id or client_secret is invalid."
                )
            response.raise_for_status()
            return response.json()["access_token"]

    async def _aget_token(self) -> str:
        """
        Asynchronously obtains a JWT token for authorization if token is not provided or has expired.

        Returns:
            str: JWT token.
        """
        if not self._client_id or not self._client_secret:
            raise ValueError(
                "client_id and client_secret must be provided to fetch JWT token."
            )

        async with httpx.AsyncClient() as client:
            response = await client.post(
                self.AUTH_URL,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data={
                    "grant_type": "client_credentials",
                    "client_id": self._client_id,
                    "client_secret": self._client_secret,
                    "scope": "openid",
                },
            )

            if response.status_code == 401:
                raise ValueError(
                    "Client Error `401 Unauthorized`: client_id or client_secret is invalid."
                )
            response.raise_for_status()
            return response.json()["access_token"]

    def _create_embedding(
        self,
        texts: List[str],
        model_name: Literal[
            "bge_base",
            "bge_large",
        ] = "bge_base",
        **kwargs: Any,
    ) -> List[List[float]]:
        """
        Creates embeddings for the provided text(s) using the Modelhub embedding model.

        Args:
            texts (List[str]): The input list of texts to generate embeddings for.
            model_name (str): Modelhub embedding model name. Defaults to "bge_base".
            kwargs (Any): Any additional arguments to be used.

        Returns:
            List[List[float]]: A list of embedding vectors for the input texts.
        """

        if not self._token or ModelhubEmbedding.is_token_expired(self._token):
            self._token = self._get_token()

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._token}",
        }

        payload = {"texts": texts, "model_name": model_name}

        # Setting verify=False, so our client doesn't check for SSL certificate.
        with httpx.Client(verify=False) as client:
            response = client.post(self.MODEL_URL, headers=headers, json=payload)
            response.raise_for_status()
            return response.json()["prediction"]

    async def acreate_embedding(
        self,
        texts: Union[str, List[str]],
        model_name: Literal[
            "bge_base",
            "bge_large",
        ] = "bge_base",
        **kwargs: Any,
    ) -> List[List[float]]:
        """
        Asynchronously creates embeddings for the provided text(s) using the Modelhub embedding model.

        Args:
            texts (Union[str, List[str]]): The input text or list of texts to generate embeddings for.
            model_name (str): Modelhub embedding model name. Defaults to "bge_base".
            kwargs (Any): Any additional arguments to be used.

        Returns:
            List[List[float]]: A list of embedding vectors for the input text(s).
        """

        if not self._token or ModelhubEmbedding.is_token_expired(self._token):
            self._token = await self._aget_token()

        if isinstance(texts, str):
            texts = [texts]

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._token}",
        }

        payload = {"texts": texts, "model_name": model_name}

        # Setting verify=False, so our client doesn't check for SSL certificate.
        async with httpx.AsyncClient(verify=False) as client:
            response = await client.post(self.MODEL_URL, headers=headers, json=payload)
            response.raise_for_status()
            return response.json()["prediction"]

==> autonomize_autorag-0.1.15-py3-none-any.whl: autorag/embedding/modelhub.py <==
"""Modelhub Embedding module implementation"""

# pylint: disable=line-too-long

import base64
import json
import time
from typing import Any, List, Literal, Optional, Union

import httpx

from autorag.embedding.base import Embedding
from autorag.utilities.logger import get_logger

logger = get_logger()


class ModelhubEmbedding(Embedding):
    """
    Modelhub Embedding implementation.

    This class provides an implementation of the Embedding abstract class
    using our internal Modelhub SDK's embedding models to generate embeddings.

    Example:
    .. code-block:: python

        from autorag.embedding.modelhub import ModelhubEmbedding

        model_name = "bge_base"
        model = ModelhubEmbedding(
            client_id="modelhub-client",
            client_secret="xxx"
        )

        model.create_embedding("Sample text")
    """

    AUTH_URL = "https://auth.sprint.autonomize.dev/realms/autonomize/protocol/openid-connect/token"
    MODEL_URL = (
        "https://clinical-llm.modelhub.sprint.autonomize.dev/v1/model/embedding/predict"
    )

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[str] = None,
    ):
        """
        Initialize the ModelhubEmbedding instance.

        Args:
            client_id (Optional[str]): The client ID for authentication.
            client_secret (Optional[str]): The client secret for authentication.
            token (Optional[str]): An existing JWT token for authentication.

        Raises:
            ValueError: If neither (client_id and client_secret) nor token is provided.
        """

        if client_id is None or client_secret is None:
            if token is None:
                raise ValueError(
                    "Either (`client_id` and `client_secret`) or `token` must be provided."
                )
            logger.warning(
                "It is recommended to provide `client_id` and `client_secret` over `token` "
                "because token gets expired within 24 hours. "
                "You can use `ModelhubEmbedding.is_token_expired(your_token)` to check if it has expired or not."
            )

        self._client_id = client_id
        self._client_secret = client_secret
        self._token = token

    @staticmethod
    def is_token_expired(token: str) -> bool:
        """
        Checks whether the provided JWT token is expired or not.

        Args:
            token (str): The JWT token to check if its expired or not.

        Returns:
            bool: True if token has expired, otherwise False.

        Raises:
            ValueError: When the token provided is not base64
        """

        # Split the token by `.` to get it in list
        b64_components = token.split(".")

        # The token should have three components: header, payload and signature.
        if len(b64_components) != 3:
            raise ValueError("Ill-formatted token provided. Please recheck your token.")

        # Divide it into, header, payload and signature,
        # but we don't need header and signature.
        _, payload_b64, _ = b64_components

        payload_json = base64.urlsafe_b64decode(payload_b64 + "==").decode("utf-8")
        payload = json.loads(payload_json)

        # Get expiry time in Unix-epoch time.
        token_expiry_time = payload["exp"]

        return token_expiry_time < int(time.time())

    def _get_token(self) -> str:
        """
        Obtains a JWT token for authorization if token is not provided or has expired.

        Returns:
            str: JWT token.
        """
        if not self._client_id or not self._client_secret:
            raise ValueError(
                "client_id and client_secret must be provided to fetch JWT token."
            )

        with httpx.Client() as client:
            response = client.post(
                self.AUTH_URL,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data={
                    "grant_type": "client_credentials",
                    "client_id": self._client_id,
                    "client_secret": self._client_secret,
                    "scope": "openid",
                },
            )

            if response.status_code == 401:
                raise ValueError(
                    "Client Error `401 Unauthorized`: client_id or client_secret is invalid."
                )
            response.raise_for_status()
            return response.json()["access_token"]

    async def _aget_token(self) -> str:
        """
        Asynchronously obtains a JWT token for authorization if token is not provided or has expired.

        Returns:
            str: JWT token.
        """
        if not self._client_id or not self._client_secret:
            raise ValueError(
                "client_id and client_secret must be provided to fetch JWT token."
            )

        async with httpx.AsyncClient() as client:
            response = await client.post(
                self.AUTH_URL,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data={
                    "grant_type": "client_credentials",
                    "client_id": self._client_id,
                    "client_secret": self._client_secret,
                    "scope": "openid",
                },
            )

            if response.status_code == 401:
                raise ValueError(
                    "Client Error `401 Unauthorized`: client_id or client_secret is invalid."
                )
            response.raise_for_status()
            return response.json()["access_token"]

    def _create_embedding(
        self,
        texts: List[str],
        model_name: Literal[
            "bge_base",
            "bge_large",
        ] = "bge_base",
        **kwargs: Any,
    ) -> List[List[float]]:
        """
        Creates embeddings for the provided text(s) using the Modelhub embedding model.

        Args:
            texts (List[str]): The input list of texts to generate embeddings for.
            model_name (str): Modelhub embedding model name. Defaults to "bge_base".
            kwargs (Any): Any additional arguments to be used.

        Returns:
            List[List[float]]: A list of embedding vectors for the input texts.
        """

        if not self._token or ModelhubEmbedding.is_token_expired(self._token):
            self._token = self._get_token()

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._token}",
        }

        payload = {"texts": texts, "model_name": model_name}

        # Setting verify=False, so our client doesn't check for SSL certificate.
        with httpx.Client(verify=False) as client:
            response = client.post(self.MODEL_URL, headers=headers, json=payload)
            response.raise_for_status()
            return response.json()["prediction"]

    async def acreate_embedding(
        self,
        texts: Union[str, List[str]],
        model_name: Literal[
            "bge_base",
            "bge_large",
        ] = "bge_base",
        **kwargs: Any,
    ) -> List[List[float]]:
        """
        Asynchronously creates embeddings for the provided text(s) using the Modelhub embedding model.

        Args:
            texts (Union[str, List[str]]): The input text or list of texts to generate embeddings for.
            model_name (str): Modelhub embedding model name. Defaults to "bge_base".
            kwargs (Any): Any additional arguments to be used.

        Returns:
            List[List[float]]: A list of embedding vectors for the input text(s).
        """

        if not self._token or ModelhubEmbedding.is_token_expired(self._token):
            self._token = await self._aget_token()

        if isinstance(texts, str):
            texts = [texts]

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._token}",
        }

        payload = {"texts": texts, "model_name": model_name}

        # Setting verify=False, so our client doesn't check for SSL certificate.
        async with httpx.AsyncClient(verify=False) as client:
            response = await client.post(self.MODEL_URL, headers=headers, json=payload)
            response.raise_for_status()
            return response.json()["prediction"]
